#!/usr/bin/perl
#
# Incident handling script v1.3
# Author: Viraj Alankar <valankar@bigfoot.com>

use Socket;
use Net::Telnet;
use POSIX qw(strftime);
use Mail::Sender;

# CONFIG SECTION
################

# NOTE: Below are default values. You can either change these or 
# create an $HOME/.incident.pl.cfg or ./incident.pl.cfg (in the current directory)
# All of the values below will be overridden by a configuration file.
# See incident.pl.cfg.sample that came with this script.

# For email contacts, always include these users of the suspect domain
@admin_emails = ( "root", "postmaster", "webmaster", "abuse", "security", "noc");

# EMAIL information
$TIMEZONE_STR = "Eastern US time - GMT -0400";
$MYDOMAIN = "mydomain.com";
$FROM_ADDR = 'Full Name <youremail@address.com>';
$SMTP_SERVER = "your.mail.server";
$ALWAYS_CC = 'someone@whatever.blah';

# TCP services to check for on suspect host
@tcp_ports = ( 21, 22, 23, 25, 79, 80, 139, 6000, 6001 );

# TCP timeout
$TCP_TIMEOUT = 5;

# Override initial string timeouts for these ports
# Basically we want to wait longer for initial string data on these ports
# if there is connection to the port within $TCP_TIMEOUT
%tcp_override_timeouts = (
	"21" => "30",
	"22" => "30",
	"23" => "30",
	"25" => "30"
);

# Reporting directory
# If not defined, will use current directory
# $REP_DIR = "/home/valankar/incident/incidents";

# Report file (in subdir of $REP_DIR)
$REP_FILE = "orig_email.txt";

# Will store email addresses of who was sent $REP_FILE (in subdir of $REP_DIR)
$SENT_TO_FILE = "orig_email_sent_to.txt";

# Where to store email followups (in subdir of $REP_DIR)
$FOLLOWUP_FILE = "email_followup.txt";

# tmp file
$TMP_FILE = "/tmp/incident.out.$$";

# Signature to append to message
# If not defined, no signature
# $SIGNATURE = "/home/valankar/.sig";

# Favorite editor
$EDITOR = "vi";

# If you change the subject format below, keep DOMAIN_INS and INCNUM_INS in the subject
# or followup processing will not work
$attack_email_subject = "Possible attack from your system - DOMAIN_INS - Incident #: INCNUM_INS";

# The actual message that will be sent to the administrators
$attack_email = <<EOF;

System Administrators, 

        We noticed what is most likely an attempted attack on our  
computers which has been logged from the address:

FQDN_INS
IP_INS

        We ask that you please look into this matter as soon as
possible.  Attached below is our logs which show evidence of the attack.
The date and times of the requests are listed below.  Please let us know 
how we can be of further assistance tracking this down. We consider this 
a very serious threat to our company systems. Thank you for your 
cooperation in this matter.

Note that all hostnames listed below are all in the $MYDOMAIN domain,
and all times are $TIMEZONE_STR.

Please keep the Subject of this message intact on all correspondence.

SIG_INS
Date of this report: DATE_INS
DESC_INS
EOF

# END CONFIG SECTION
####################

sub catch_ctrlc {
	my $signame = shift;

	print("Removed $curr_rep_dir\n") if (-d $curr_rep_dir && rmdir($curr_rep_dir));
	unlink($TMP_FILE);
	die("Quitting on SIG$signame.\n");
}

# Find what seems to be an attack
# This is determined by finding out the most number of occurrences of an IP
# address in the input
sub get_suspects {
	my $max;
	my %ips;

	foreach (@_) {
		if (/(\d+\.\d+\.\d+\.\d+)/) {
			# Ignore IPs that syslog appears to be from?
			# next if (/^\S+\s+\S+\s+\S+\s+$1/);
			$ips{$1}++;
			$max = $1 if ($ips{$1} > $ips{$max});
		}
	}
	return ($max);
}

sub reverse_lookup {
	return (gethostbyaddr(inet_aton($_[0]), AF_INET));
}

sub isolate_input {
	my $rev = pop(@_);
	my $ip = pop(@_);
	my @result;

	foreach (@_) {
		push(@result, $_) if (/$ip/ || ($rev && /$rev/));
	}

	return @result;
}

# Try to get whois information on domains from www.allwhois.com
# TODO: use HTTP module instead of wget
# TODO: maybe use an auto-forwarder like RH7.0 whois
sub get_allwhois {
	my @result;
	my $start;
	my $work_addr = $_[0];

	# Example hostname: reality.cse.fau.edu.co
	# We have to try a whois on all extensions until we get info:
	#   cse.fau.edu.co
	#   fau.edu.co
	#   edu.co
	# 
	# We really will not try a whois on 'co' only, however I have seen domains
	# like this (www.ni for example)

	while (!@result && $work_addr =~ /.+\..+/) {
		print("WHOIS $work_addr...");

		@webout = `wget -q -O - http://www.allwhois.com/cgi-bin/allwhois5.cgi?dn=$work_addr`;

		# Tricky, parse Allwhois output
		$start = 0;
		foreach (@webout) {
			push(@result, $_) if ($start);
			last if (@result && $result[$#result] =~ s/<\/TEXTAREA.*//);
			$start = 1 if ($start != 1 && /<TEXTAREA NAME=description/);
		}

		if (grep(/Error:/, @result)) {
			print("failed.\n");
			@result = ();
		} else {
			print("succeeded.\n");
		}

		# Strip out first prefix for next iteration
		$work_addr =~ s/.*?\.//;
	}

	return @result;
}
	
# ARIN may redirect us to other netblocks which makes this difficult
sub get_arin_whois {
	my @result;
	my $work_ip = $_[0];
	my $orig_ip = $_[0];

	until (@result) {
		print("WHOIS $work_ip...");
		@result = `whois $work_ip\@whois.arin.net`;

		# Take care of RIPE redirections
		if (grep (/whois.ripe.net/, @result)) {
			print("ARIN redirect to RIPE...");
			@result = `whois $work_ip\@whois.ripe.net`;
		} elsif (grep(/whois.apnic.net/, @result)) {
			print("ARIN redirect to APNIC...");
			@result = `whois $work_ip\@whois.apnic.net`;
		} elsif (grep(/To single out one record/, @result)) {
			my %netblocks;
			my $diff;
			my $min = -1;
			my $min_netblock;
			my $current_netblock;

			foreach (@result) {
				if (/\(([A-Z]+.*)\)/) {
					$current_netblock = $1;
				}
				next if (!/(\d+\.\d+\.\d+\.\d+ - \d+\.\d+\.\d+\.\d+)/);
				$netblocks{$current_netblock} = "$1,$2";
			}

			# Find which netblock holds our IP to the closest range
			foreach (keys %netblocks) {
				$diff = find_netblock_diff($orig_ip, split(/,/, $netblocks{$_}));
				if ($min == -1 || $diff < $min) {
					$min = $diff;
					$min_netblock = $_;
				}
			}
			print("ARIN redirect to netblock $min_netblock\n");
			$work_ip = "!$min_netblock";
			@result = ();
		}
		# Sleep a little bit in case of infinite loop
		sleep(1);
	}

	print("succeeded.\n");
	return @result;
}

# Find the netblock with the closes beginning IP (end doesn't matter)
sub find_netblock_diff {
	my $ip = $_[0];
	my $begin = $_[1];
	my $end = $_[2];
	my $diff = 0;
	
	@begspl = split(/\./, $begin);
	@ipspl = split(/\./, $ip);

	for (my $i = 0; $i < @ipspl; $i++) {
		$diff += $ipspl[$i] - $begspl[$i];
	}
	return $diff;
}

sub get_whois {
	my @result;

	if ($_[0] =~ /^\d+\.\d+\.\d+\.\d+$/) {
		# It's an IP, so use ARIN
		@result = get_arin_whois($_[0]);
	} else {
		@result = get_allwhois($_[0]);
	}

	return @result;
}

# Return a list of possible email addresses
sub get_emails {
	my %emails;
	my $email;
	
	foreach (@_) {
		if (/(\S+\@[A-Za-z]+\S*\.\S+)/) {
			($email = lc($1)) =~ s/mailto://g;
			$email =~ s/[^\@\w\.\-]//g;
			$emails{$email}++;
		}
	}
	return (sort keys %emails);
}

sub host_probe {
	my @result;
	my $conn_status;
	my @answ_ports;

	foreach (@tcp_ports) {
		$conn_status = tcp_connect($_[0], $_);
		if ($conn_status) {
			push(@result, $conn_status);
			push(@answ_ports, $_);
		}
	}

	# handle any operations for special ports
	foreach (@answ_ports) {
		push(@result, port_special($_[0], $_, $_[1]));
	}
	return @result;
}

sub port_special {
	my $ip = $_[0];
	my $port = $_[1];
	my $storage = $_[2];
	my $result;

	if ($port == 139) {
		$result .= probe_smbclient($ip);
	} elsif ($port == 79) {
		$result .= probe_finger($ip);
	} elsif ($port == 80) {
		$result .= probe_http($ip);
	} elsif ($port >= 6000 && $port <= 6010) {
		$result .= probe_xwd($ip, $port, $storage);
	}

	return $result;
}

sub probe_xwd {
	my $result;
	my $dispnum;
	my $cmd;
	my $show_str;

	$dispnum = $_[1] - 6000;

	$show_str = "XWD SCREENDUMP OF $_[0]:$dispnum";
	print $show_str;

	$cmd = "xwd -display $_[0]:$dispnum -out $_[2]/$_[0].xwd -root 2>&1";

	$result = "$show_str$cmd\n";
	$result =~ s/-out .*\.xwd/-out $_[0].xwd/;
	$result .= `$cmd` . "\n";

	if ($? >> 8) {
		$result = "";
		print(" failed\n");
	} else {
		print(" succeeded\n");
	}
	return $result;
}
	
sub probe_smbclient {
	my $result;
	my $show_str;

	$show_str = "SMBCLIENT SHARE LIST OF $_[0]\n";
	print $show_str;

	$result = `smbclient -N -L $_[0] 2>/dev/null`;
	$result =~ s/^[Aa]dded interface.*//gm;
	$result =~ s/^session request to .* failed.*//gm;
	$result =~ s/^\n//gm;

	$result = "$show_str$result" if ($result);
	return $result;
}

sub probe_finger {
	my $output;
	my $show_str;

	$show_str = "FINGER \@$_[0]\n";
	print $show_str;

	$output = `finger \@$_[0] 2>&1`;
	
	$output = "$show_str$output\n" if ($output);
	return $output;
}

sub probe_http {
	my $output;
	my $show_str;

	$show_str = "HTTP GET / TO $_[0]\n";
	print $show_str;
	$output = `wget -q -O - $_[0] 2>&1`; 
	$output = "$show_str$output\n\n" if ($output);
	return $output;
}

sub tcp_connect {
	my $ip = $_[0];
	my $port = $_[1];
	my $line;
	my $result;
	my $obj;
	my $show_str;
	my $timeout;

	$show_str = "TCP CONNECT TO $ip:$port";
	print $show_str;

	$result = "$show_str succeeded\n";

	$obj = new Net::Telnet (Errmode => "return");
	$obj->open(Host => $ip, Port => $port, Timeout => $TCP_TIMEOUT);

	$timeout = $tcp_override_timeouts{$port};
	$timeout = $TCP_TIMEOUT if (!$timeout);
	while (!$obj->errmsg && ($line = $obj->get(Timeout => $timeout))) {
		$result .= $line;
		$timeout = $TCP_TIMEOUT;
	}

	$obj->close();

	if ($obj->errmsg() && $obj->errmsg !~ /^read timed-out/) {
		print(" failed\n");
		$result = "";
	} else {
		$result .= "\n" if ($result !~ /\n$/);
		$result .= "Connection closed.\n\n";
		print(" succeeded\n");
	}
	return $result;
}

sub get_y_or_n {
	my $answer;

	print $_[0];
	if (($answer = <STDIN>) =~ /^y/i) {
		return 1;
	}
	return 0;
}

sub get_description {
	my @desc;

	if (get_y_or_n("Add description of attack to message? ")) {
		print("Enter description (ctrl-d on blank line to end):\n");
		@desc = <STDIN>;
		chop(@desc);
	}
	return @desc;
}

sub dig_domain {
	my $result;
	my $work_ip = $_[0];

	while (!$result && $work_ip =~ /\d+\.\d+/) {
		@digout = `dig -x $work_ip`;
		$look_for = join('.', reverse(split(/\./, $work_ip))) . ".in-addr.arpa.";
		if (@found = grep(/^$look_for/i, @digout)) {
			if ($found[0] =~ /\s+SOA\s+(\S+)/) {
				# It's an authoritative line
				$result = $1;
			} else {
				my @tmparr = split(/\s+/, $found[0]);
				$result = $tmparr[$#tmparr];
			}
			$result =~ s/\.$//;
			return($result);
		}

		# Strip off end part of IP and try again
		$work_ip =~ s/\.\d+?$//;
	}

	return "";
}

sub email_pipe_followup {
	my $lookfor;
	my $input;
	my $domain;
	my $incnum;
	my $dest_file;
	my $incopy;
	
	print("Analyzing piped input as email followup...\n");
	$lookfor = $attack_email_subject;
	$lookfor =~ s/DOMAIN_INS/\(\\S*)/;
	$lookfor =~ s/INCNUM_INS/\(\\d\{14,\}\)/;
	$lookfor =~ s/\s+/\\s+/g;
	$input = join('', <STDIN>);
	($incopy = $input) =~ s/^\s*>\s*//gm;
	if ($incopy !~ /$lookfor/m) {
		print STDERR ("Cannot determine what input email is in reference to. Quitting.\n");
		print STDERR ("Regular expression was:\n$lookfor\n");
		return 1;
	} else {
		$domain = $1;
		$incnum = $2;
	}

	$input = "From XXX " . strftime("%a %b %e %H:%M:%S %Z %Y", localtime) . "\n" . $input 
		if ($input !~ /^From /);
	$dest_file = "$REP_DIR/$domain/$incnum/$FOLLOWUP_FILE";
	if (!open(OFILE, ">>$dest_file")) {
		print STDERR ("Cannot create $dest_file: $!\n");
		return 1;
	}
	print OFILE $input, "\n";
	close(OFILE);
	print("Wrote email to $dest_file\n");
	return 0;
}

sub load_configuration {
	my $hconf = $ENV{HOME} . "/.incident.pl.cfg";
	my $cconf = "./incident.pl.cfg";

	if (-e $hconf && -e $cconf) {
		die("Configuration found in both $hconf and $cconf. Please pick one to use and delete the other\n");
	}

	if (-e $hconf) {
		require $hconf;
	} elsif (-e $cconf) {
		require $cconf;
	}
}

#### Start

$SIG{INT} = \&catch_ctrlc;
$| = 1;

load_configuration();

# Email piping
if ($ARGV[0] eq "-e") {
	exit(email_pipe_followup());
}

print("Analyzing input...\n");
# Take care of any isolations specified
if (@ARGV == 2) {
	$isolate = $ARGV[1];
	$isolate =~ s/\./\\\./g;
	pop(@ARGV);
}

@input = <>;

if ($isolate) {
	@suspinput = grep(/$isolate/, @input);
} else {
	@suspinput = @input;
}

@suspects = get_suspects(@suspinput);

$REP_DIR = `pwd` if (!$REP_DIR);
chomp($REP_DIR);

foreach (@suspects) {
	print("Attack appears to be from $_\n");
	print("Reverse lookup: ");
	printf("%s\n", ($rev = reverse_lookup($_)) ? $rev : "failed");
	$rev = "" if ($rev eq $_);

	print("Isolating input for $_" . ($rev ? " and $rev" : "") . "\n");
	@newlogs = isolate_input(@input, $_, $rev);

	$start_time = time();
	$now_string = strftime("%a %b %e %H:%M:%S %Z %Y", localtime($start_time));
	$dirname = strftime("%Y%m%d%H%M%S", localtime($start_time));

	print("\nInformation Gathering\n\n$now_string\n\n");

	@whois = get_whois($_);
	@whois_rev = get_whois($rev) if ($rev);

	# If reverse lookups fail, dig for the domain
	if (!$whois_rev[0]) {
		print("Trying to figure out domain name after failed reverse lookup.\n");
		$domain = dig_domain($_);
		if ($domain) {
			print("Possible candidate: $domain\n");
			@whois_rev = get_whois($domain);
		} else {
			print("Failed to determine any domain.\n");
		}
	}
	chop($domain = $whois_rev[0]);

	@emails = get_emails(@whois, @whois_rev);

	die("No email contacts found. Quitting.\n") if (!@emails);

	# If domain is still unknown, assume it is domain of the first email
	# returned by ARIN
	# TODO: This is probably the wrong thing to do.
	(undef, $domain) = split(/\@/, $emails[0]) if (!$domain);

	$domain = lc($domain);

	push(@emails, "$_\@$domain") foreach (@admin_emails);

	$curr_rep_dir = "$REP_DIR/$domain/$dirname";
	system("mkdir -p $curr_rep_dir");

	@host_info = host_probe($_, $curr_rep_dir);

	$out_msg = $attack_email;
	@desc = get_description(@newlogs);
	$tmp_desc = "\nDescription of this attack:\n" . join("\n", @desc) .
		"\n" if (@desc);
	$out_msg =~ s/DESC_INS/$tmp_desc/;
	$out_msg =~ s/IP_INS/$_/g;
	$out_msg =~ s/FQDN_INS/$rev/g;
	$out_msg =~ s/DATE_INS/$now_string/g;

	$msg_subj = $attack_email_subject;
	$msg_subj =~ s/DOMAIN_INS/$domain/g;
	$msg_subj =~ s/INCNUM_INS/$dirname/g;

	$sig = "";
	$sig = `cat $SIGNATURE` if (-e $SIGNATURE);
	$out_msg =~ s/SIG_INS/$sig/g;

	open(FOUT, ">$curr_rep_dir/$REP_FILE") or die("Cannot create: $!\n");
	print FOUT $out_msg;
	print FOUT "Logs of attack:\n", @newlogs, "\n";
	print FOUT "Name: $rev\n" if ($rev);
	print FOUT "Address: $_\n\n";
	print FOUT @whois, "\n";
	print FOUT @whois_rev;
	print FOUT @host_info;
	close(FOUT);
	print("Email for $_ written to $curr_rep_dir/$REP_FILE\n");

	if (get_y_or_n("\nView/edit message? ")) {
		system("$EDITOR $curr_rep_dir/$REP_FILE");
	}

	open(FOUT, ">$curr_rep_dir/$SENT_TO_FILE") or warn("Cannot create: $!\n");
	print ("\nPossible admin contacts:\n");
	$mailto = "";
	foreach (@emails) {
		$mailto .= "$_,";
		print ("$_,\n");
		print FOUT ("$_\n");
	}
	close(FOUT);
	$mailto =~ s/,$//;

	if (get_y_or_n("\nSend message? ")) {
		my $body = `cat $curr_rep_dir/$REP_FILE`;
		my $sender = new Mail::Sender( { from => $FROM_ADDR,
			smtp => $SMTP_SERVER } ) or warn("$Mail::Sender::Error\n");
		$sender->MailMsg( { to => $mailto, subject => $msg_subj, msg => $body,
			cc => $ALWAYS_CC } ) or warn("$Mail::Sender::Error\n");
	}
}

unlink($TMP_FILE);
